#!/usr/bin/perl

#    This distribution makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman> email: sanjayjaiman@yahoo.com\n";
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

my $globalTokenCount = 0;

# Each node contains an array of the following
# 0. type - [val/array]
# 1. Val  - if type == val, ref to list of values (array ref)
#           if type == array, array of same type as this one (hash ref)
# 2. Num  - Number of elements in Val if type == val/ Number of children if type == array
# 3. Name
# 4. Level
%globalVariableTree;

$numGlobalVariables;

@structureNames;
# structures hash - Key = structure name
#                 - Value = ref to array containing
#                         T - Type of structure
#                         Count - Num of fields
#                         Fields name array ref
#                         Fields type array ref
#                         Fields T Val array ref
%structureFields;
@functionNames;
# functions hash - Key = function name
#                - Value = ref to array containing
#                         T - Type of function
#                         count of input array elements
#                         Input name array ref
#                         count of input array elements
#                         Output name array ref
%functionFields;

@supportedTypes = ("int", "uint", "string", "float", "ipaddress");

sub getNext {
  my ($nameRef, $hashRef, $num) = @_;
  $$nameRef = "";
  foreach $key (keys(%$hashRef)) {
    my $order = $$hashRef{$key}->[5];
    if ($order == $num) {
      $$nameRef = $key;
      return 1;
    }
  }
  return 0;
}
sub printVars {
  my ($name, $hashRef, $numElements) = @_;
  my $len = @hashExp;
  my $i = 0;
#  print "Num elements in $name = $numElements\n";
#  foreach $key (keys(%$hashRef)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $key = "";
    if (getNext(\$key, $hashRef, $hashCount) == 0) {
      last;
    }
    my $name = $key;
    my $type  = $$hashRef{$key}->[0];
    my $name2 = $$hashRef{$key}->[3];
    my $level = $$hashRef{$key}->[4];
    my $orderNum = $$hashRef{$key}->[5];
    my $Ref = $$hashRef{$key}->[1];
    my $pre = "";

    for (my $j = 0; $j < $level; $j++) {
      $pre = $pre."\t";
    }
    print $pre."name = $name2; type = $type; Level = $level; Order = $orderNum;";
    if ($type eq "Val") {
      my $arrayCount = $$hashRef{$key}->[2];
      print " Val = ";
      for (my $i = 0; $i < $arrayCount; $i++) {
	print "$Ref->[$i] ";
      }
      print "\n";
    }
    if ($type eq "Array") {
      my $arrayCount = $$hashRef{$key}->[2];
      print "\n";
      printVars($name, $Ref, $arrayCount);
    }
  }
}

sub tokenize {
  my ($arrayRef, $tokenCountRef, $line) = @_;

  while ($line ne "") {
    $line =~ s/\s*//;  # Get rid of leading spaces
    my $tok = $line;
    # Look for the next space and replace the string starting at that point
    # till the end of the line with nothing
    $tok =~ s/\s.*//;
    $line =~ s/$tok//;  # Skip over the token
    $$arrayRef[$$tokenCountRef++] = $tok;
  }
}

sub isValToken {
  my ($tokensRef, $whichone) = @_;
  my $retVal = 0;
  #printf "isValToken:: checking $$tokensRef[$whichone]\n";
  if ($$tokensRef[$whichone] =~ /:/) {
    $retVal = 1;
  }
  #printf "isValToken:: returning $retVal\n";
  return $retVal;
}

sub isArrayToken {
  my ($tokensRef, $whichone) = @_;
  my $retVal = 0;
  #printf "isArrayToken:: checking $$tokensRef[$whichone]\n";
  if ($$tokensRef[$whichone] =~ /[^:\{\}]/) {
  if ($$tokensRef[$whichone+1] eq "{") {
      $retVal = 1;
    }
  }
  #printf "isArrayToken:: returning $retVal\n";
  return $retVal;
}

sub valDone {
  my ($tokensRef, $whichone) = @_;
  my $retVal = 0;
  if (isValToken($tokensRef, $whichone) || isArrayToken($tokensRef, $whichone) ||
      ($$tokensRef[$whichone] =~ /\}/) || $whichone >= $globalTokenCount) {
    $retVal = 1;
  }
  #printf "valDone:: $$tokensRef[$whichone] :: returning $retVal\n";
  return $retVal;
}

sub process {
  my ($tokensRef, $tokenPtrRef, $hashRef, $levelRef, $order) = @_;
  my $j = 0;
  my $retVal = 1;
  if (isValToken($tokensRef, $$tokenPtrRef)) {
    # Val token found
    my $refToNewArray;
    my $valTokenCount = 0;
    my $name = $$tokensRef[$$tokenPtrRef++];

    while (!valDone($tokensRef, $$tokenPtrRef)) {
      $refToNewArray->[1][$valTokenCount++] = $$tokensRef[$$tokenPtrRef++];
    }
    $name =~ s/://;
    $refToNewArray->[0] = "Val";
    $refToNewArray->[2] = $valTokenCount;
    $refToNewArray->[3] = $name;
    $refToNewArray->[4] = $$levelRef;
    $refToNewArray->[5] = $$order++;

#    print "Val tokens of $name = ";
#    for (my $i = 0; $i < $valTokenCount; $i++) {
#      print "$refToNewArray->[1][$i] ";
#    }
#    print " :: Count = $valTokenCount\n";
    $$hashRef{$name} = $refToNewArray;
    my $arrayCount = $$hashRef{$name}->[2];
    my $arrayRef = $$hashRef{$name}->[1];
#    print "*** Adding :: Type = $$hashRef{$name}->[0] :: Name = $$hashRef{$name}->[3] :: Level = $$hashRef{$name}->[4] :: Count = $arrayCount :: Tokens :: ";
#    for (my $i = 0; $i < $arrayCount; $i++) {
#      print "$arrayRef->[$i] ";
#    }
#    print "\n";
  }
  elsif (isArrayToken($tokensRef, $$tokenPtrRef)) {
    my $name = $$tokensRef[$$tokenPtrRef++];
    # skip over {
    $retVal = 0;
    if ($$tokensRef[$$tokenPtrRef] eq "{") {
#      print "Start of structure\n";
      $$tokenPtrRef++;
      my $refToNewArray;
      my $tempLevel = $$levelRef;
      $refToNewArray->[0] = "Array";
      $refToNewArray->[3] = $name;
      $refToNewArray->[4] = $$levelRef;
      $refToNewArray->[5] = $$order++;
      $$levelRef++;
      $$hashRef{$name} = $refToNewArray;
#      print "--- Adding :: Type = $$hashRef{$name}->[0] :: Name = $$hashRef{$name}->[3] :: Level = $$hashRef{$name}->[4] :: Count = $arrayCount :: Next Level ::\n";
      my %temp;
      my $numChildren = 0;
      my $j = 0;
      while ($$levelRef > $tempLevel) {
	if (process($tokensRef, $tokenPtrRef, \%temp, $levelRef, \$j)) {
	  $$hashRef{$name}->[1] = \%temp;
	  $numChildren++;
	}
      }
      $refToNewArray->[2] = $numChildren - 1;  # Number of tokens (values) (-1 because "}" got counted
      $retVal = 1;
    }
    else {
      printf "ERROR::Start of structure not found\n";
      $retVal = 0;
    }
  }
  elsif ($$tokensRef[$$tokenPtrRef] eq "{") {
    print "ERROR::Skipping over $tokensRef[$$tokenPtrRef]\n";
    $$tokenPtrRef++;
  }
  elsif ($$tokensRef[$$tokenPtrRef] eq "}") {
#    print "End of structure\n";
    $$tokenPtrRef++;
    $$levelRef--;
    $retVal = 1;
  }
  elsif ($$tokenPtrRef >= $globalTokenCount) {
    print "End of file\n";
    $retVal = 0;
  }
  return $retVal;
}

sub parseFile {
  my ($input, $output) = @_;

  my @tokens;
  while ($line = <$input>) {
#    print "Line::$line";
    if ($line =~ /^\#/) {
      next;
    }
    $line =~ s/\n//;
    tokenize(\@tokens, \$globalTokenCount, $line);
  }
  print $output "Tokens:: @tokens\n\n";
  my $i = 0;
  my $level = 0;

  my $j = 0;
  while (process(\@tokens, \$i, \%globalVariableTree, \$level, \$j)) {
    $numGlobalVariables++;
  }
}

sub getVals {
  my ($name, $inputArrayRef, $arraylen, $outputArrayRef) = @_;
  my $i = 0;
  for ($i = 0; $i < $arraylen; $i++) {
    $$outputArrayRef[$i] = $$inputArrayRef[$i];
  }
}

sub getChildren {
  my ($parent, $hashRef, $outputArrayRef) = @_;
  my $i = 0;
#  print "$parent :: Key/Vals\n";
#  foreach $key (keys(%$hashRef)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $key = "";
    if (getNext(\$key, $hashRef, $hashCount) == 0) {
      last;
    }
#    print "$key/$$hashRef{$key} ";
    my $name  = $$hashRef{$key}->[3];
    my $len  = $$hashRef{$key}->[2];
    my $Ref = $$hashRef{$key}->[1];
    $$outputArrayRef[$i++] = "$name";
  }
#  print "\n";
}

sub getValues {
  my ($inputArrayRef, $outputArrayRef, $typeRef) = @_;
  my $arraylen = @$inputArrayRef;
  my $hashRef = \%globalVariableTree;
  my $hashArrayLen = $numGlobalVariables;
  $$typeRef = "Array";
  if ($arraylen == 0) {  #root
    my $name = "root";
    getChildren($name, $hashRef, $outputArrayRef);
    return;
  }
  my $i = 0;
  my $lastmatch;
  my $lastHashMatchRef = $hashRef;
  for (my $i = 0; $i < $arraylen; $i++) {
    my $str = $$inputArrayRef[$i];
    my $arrayRef = $$lastHashMatchRef{$str};
    if ($arrayRef ne "") {
      if ($$arrayRef[0] eq "Val") {
	$$typeRef = "Val";
	getVals($str, $$arrayRef[1], $$arrayRef[2], $outputArrayRef);
	return;
      }
      else { # hash
	$lastHashMatchRef = $$arrayRef[1];
	$lastmatch = $arrayRef;
      }
    }
    else {
      return;
    }
  }
  getChildren($$lastmatch[3], $$lastmatch[1], $outputArrayRef);
  return;
}

sub printArray {
  my $len = @_;
  for (my $i = 0; $i < $len; $i++) {
    print "@_[$i], ";
  }
  print "\n";
}


sub doUserIo {

  print "Hello user - start browsing.....\n";
  my $userInput;
  while ($userInput ne "quit") {
    print "Enter dotted object list!\n>";
    $userInput = <STDIN>;
    $userInput =~ s/\s*//;
    $userInput =~ s/\s.*//;
#    print "You typed $userInput\n";
    my @inputArray = split(/\./, $userInput);
    my @outputArray;
    my $type;
#    print "input Array - @inputArray\n";
    my $len = @inputArray;
    getValues(\@inputArray, \@outputArray, \$type);
    if ($type eq "Val") {
      print "$type :: Values:: ";
      printArray(@outputArray);
    }
    else {
      print "$type :: Children:: ";
      printArray(@outputArray);
    }
  }
}

sub buildTLVStructures {
  @inputArray = ("structures");
  my $type = "None";
  getValues(\@inputArray, \@structureNames, \$type);
  if ($type ne "Array") {
    print("EEROR 1\n");
    exit;
  }
  for (my $i = 0; $i < @structureNames; $i++) {
    @inputArray = ("structures", $structureNames[$i]);
    my @flds;
    getValues(\@inputArray, \@flds, \$type);
    if ($type ne "Array") {
      print("EEROR 2\n");
      exit;
    }
    my $fieldTValArrayRef;
    my $fieldTypeArrayRef;
    my $fieldNameArrayRef;
    my $count = 0;
    for (my $j = 0; $j < @flds; $j++) {
      if ($flds[$j] ne "T") {
	my @inputArray = ("structures", $structureNames[$i], $flds[$j]);
	my @vals;
	getValues(\@inputArray, \@vals, \$type);
	$fieldNameArrayRef->[$count] = $flds[$j];
	my $found = 0;
	for (my $j = 0; $j < @supportedTypes; $j++) {
	  if ($vals[0] eq $supportedTypes[$j]) {
	    $found = 1;
	  }
	}
	if (!$found) {
	  $fieldTypeArrayRef->[$count] = "struct $vals[0]"; # assume struct
	}
	else {
	  $fieldTypeArrayRef->[$count] = $vals[0];
	}
	$fieldTValArrayRef->[$count] = $vals[1];
	$count++;
      }
    }
    @inputArray = ("structures", $structureNames[$i], "T");
    my @vals;
    getValues(\@inputArray, \@vals, \$type);
    my $refToNewArray;
    $refToNewArray->[0] = $vals[0];
    $refToNewArray->[1] = $count;
    $refToNewArray->[2] = $fieldNameArrayRef;
    $refToNewArray->[3] = $fieldTypeArrayRef;
    $refToNewArray->[4] = $fieldTValArrayRef;
    $refToNewArray->[5] = $i;
    $structureFields{$structureNames[$i]} = $refToNewArray;
  }

  @inputArray = ("functions");
  my $type = "None";
  getValues(\@inputArray, \@functionNames, \$type);
  if ($type ne "Array") {
    print("EEROR 1\n");
    exit;
  }
  for (my $i = 0; $i < @functionNames; $i++) {
    my $functionName = $functionNames[$i];
    @inputArray = ("functions", $functionName, "T");
    my @vals;
    getValues(\@inputArray, \@vals, \$type);
    my $refToNewArray;
    $refToNewArray->[0] = $vals[0];
    my $refToOutputArray;
    my $inputArrayCount = 0;
    my $outputArrayCount = 0;
    @inputArray = ("functions", $functionName, "input");
    my @IArray;
    getValues(\@inputArray, \@IArray, \$type);
    @inputArray = ("functions", $functionName, "output");
    my @OArray;
    getValues(\@inputArray, \@OArray, \$type);
    my $lastIArrayIndex = $#IArray;
    $refToNewArray->[1] =  $lastIArrayIndex + 1;  #Last index + 1
    $refToNewArray->[2] = \@IArray;
    $refToNewArray->[3] = $#OArray + 1;  #Last index + 1
    $refToNewArray->[4] = \@OArray;
    $refToNewArray->[5] = $i;
    $functionFields{$functionName} = $refToNewArray;
  }
}

sub getName {
  @inputArray = ("name");
  my @vals;
  getValues(\@inputArray, \@vals, \$type);
  return $vals[0];
}

sub getFileName {
  return getName()."ConnBase";
}

sub getPort {
  @inputArray = ("port");
  my @vals;
  getValues(\@inputArray, \@vals, \$type);
  return $vals[0];
}

sub getStructures {
  return \@structureNames;
}

sub getStructType {
  my ($structName) = @_;
  my $fieldsRef = $structureFields{$structName};
  return $fieldsRef->[0];
}

sub getStructNumFields {
  my ($structName) = @_;
  my $fieldsRef = $structureFields{$structName};
  return $fieldsRef->[1];
}

sub getFields {
  my ($structName) = @_;
  my $fieldsRef = $structureFields{$structName};
  return $fieldsRef->[2];
}

sub getFieldTypes {
  my ($structName) = @_;
  my $fieldsRef = $structureFields{$structName};
  return $fieldsRef->[3];
}

sub getFieldTVals {
  my ($structName) = @_;
  my $fieldsRef = $structureFields{$structName};
  return $fieldsRef->[4];
}

sub getNumOfTypeInStruct {
  my ($structName, $type) = @_;
  my $numFields = getStructNumFields($structName);
  my $fieldNameArrayRef = getFields($structName);
  my $fieldTypeArrayRef = getFieldTypes($structName);
  my $fieldTValArrayRef = getFieldTVals($structName);
  my $count = 0;
  for (my $i = 0; $i < $numFields; $i++) {
    if ($$fieldTypeArrayRef[$i] eq $type) {
      $count++;
    }
  }
  return $count;
}


sub getNumOfStructsInStruct {
  my ($structName, $type) = @_;
  my $numFields = getStructNumFields($structName);
  my $fieldNameArrayRef = getFields($structName);
  my $fieldTypeArrayRef = getFieldTypes($structName);
  my $fieldTValArrayRef = getFieldTVals($structName);
  my $count = 0;
  for (my $i = 0; $i < $numFields; $i++) {
    if ($$fieldTypeArrayRef[$i] =~ /struct/) {
      $count++;
    }
  }
  return $count;
}

sub getFuncType {
  my ($funcName) = @_;
  my $fieldsRef = $functionFields{$funcName};
  return $fieldsRef->[0];
}

sub getFuncInputArrayCount {
  my ($funcName) = @_;
  my $fieldsRef = $functionFields{$funcName};
  return $fieldsRef->[1];
}

sub getFuncInputArray {
  my ($funcName) = @_;
  my $fieldsRef = $functionFields{$funcName};
  return $fieldsRef->[2];
}

sub getFuncOutputArrayCount {
  my ($funcName) = @_;
  my $fieldsRef = $functionFields{$funcName};
  return $fieldsRef->[3];
}

sub getFuncOutputArray {
  my ($funcName) = @_;
  my $fieldsRef = $functionFields{$funcName};
  return $fieldsRef->[4];
}

sub printFunctions {
  print "\n\n   PRINTING FUNCTIONS\n";
#  foreach $key (keys(%functionFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $key = "";
    if (getNext(\$key, \%functionFields, $hashCount) == 0) {
      last;
    }
    my $fieldsRef = $hashRef{$key};
    my $Type = getFuncType($key);
    my $inputArrayCount = getFuncInputArrayCount($key);
    my $inputArrayRef = getFuncInputArray($key);
    my $outputArrayCount = getFuncOutputArrayCount($key);
    my $outputArrayRef = getFuncOutputArray($key);

    print ("Name = $key, Type = $Type\n\tInputArray::(num params = $inputArrayCount)::");
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      print "$$inputArrayRef[$i] ";
    }
    print "\n\tOutputArray::(num params = $outputArrayCount)::";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      print "$$outputArrayRef[$i] ";
    }
    print "\n";
  }
}

sub printStructures {
  print "\n\n   PRINTING STRUCTURES\n";
#  foreach $key (keys(%structureFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $key = "";
    if (getNext(\$key, \%structureFields, $hashCount) == 0) {
      last;
    }
    my $fieldsRef = $structureFields{$key};
    my $structType = getStructType($key);
    my $numFields = getStructNumFields($key);
    my $fieldNameArrayRef = getFields($key);
    my $fieldTypeArrayRef = getFieldTypes($key);
    my $fieldTValArrayRef = getFieldTVals($key);

    print ("Name = $key, Type = $structType, Num = $numFields, Fields ::\n");
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $fieldType = $$fieldTypeArrayRef[$j];
      my $fieldTVal = $$fieldTValArrayRef[$j];
      print "\tName - $fieldName : Type - $fieldType : T - $fieldTVal\n";
    }
    print("\n");
  }
}

sub printStructures1 {
  print "\n\n   PRINTING STRUCTURES\n";
#  foreach $key (keys(%structureFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $key = "";
    if (getNext(\$key, \%structureFields, $hashCount) == 0) {
      last;
    }
    my $fieldsRef = $structureFields{$key};
    my $numFields = $fieldsRef->[1];
    my $fieldNameArrayRef = $fieldsRef->[2];
    my $fieldTypeArrayRef = $fieldsRef->[3];
    my $fieldTValArrayRef = $fieldsRef->[4];

    print ("Name = $key, Type = $fieldsRef->[0], Num = $numFields, Fields ::\n");
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $fieldNameArrayRef->[$j];
      my $fieldType = $fieldTypeArrayRef->[$j];
      my $fieldTVal = $fieldTValArrayRef->[$j];
      print "\tName - $fieldName : Type - $fieldType : T - $fieldTVal\n";
    }
    print("\n");
  }
}

sub printConnectionBaseHFile {
  my $fileName = getFileName();
  my $className = getFileName();
  my $UfileName =  "\U$fileName\E";
  $fileName .= ".h";
  print "Wrinting file $fileName\n";
  open(SESAME, ">".$fileName) or die "Could not open output file $fileName";
  my $str = "";
  $str .= "/*\n";
  $str .= "  ############################################################################\n";
  $str .= "  #\n";
  $str .= "  #  $fileName\n";
  $str .= "  #\n";
  $str .= "  #    This is a generated file.  The functions defined in this file are primarily \n";
  $str .= "  #    used for serializing and de-serializing data structures generated from the IDL.\n";
  $str .= "  #\n";
  $str .= "  #    This distribution makes asynchronous client/server communication possible \n";
  $str .= "  #    using TLVs (Type Length Value). It achieves that letting the user define\n";
  $str .= "  #    an IDL which serves as input to a code generator.  The code generated is linked in\n";
  $str .= "  #    the client and the server program.\n";
  $str .= "  #\n";
  $str .= "  #    Copyright (C) <2005>  <Sanjay Jaiman> email: sanjayjaiman\@yahoo.com\n";
  $str .= "  #\n";
  $str .= "  #    This program is free software; you can redistribute it and/or modify\n";
  $str .= "  #    it under the terms of the GNU General Public License as published by\n";
  $str .= "  #    the Free Software Foundation; either version 2 of the License, or\n";
  $str .= "  #    (at your option) any later version.\n";
  $str .= "  #\n";
  $str .= "  #    This program is distributed in the hope that it will be useful,\n";
  $str .= "  #    but WITHOUT ANY WARRANTY; without even the implied warranty of\n";
  $str .= "  #    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n";
  $str .= "  #    GNU General Public License for more details.\n";
  $str .= "  #\n";
  $str .= "  #    You should have received a copy of the GNU General Public License\n";
  $str .= "  #    along with this program; if not, write to the Free Software\n";
  $str .= "  #    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n";
  $str .= "  #\n";
  $str .= "  ##########################################################################\n";
  $str .= "*/\n";
  $str .= "\n";
  $str .= "#ifndef _".$UfileName."_H_\n";
  $str .= "#define _".$UfileName."_H_\n";
  $str .= "\n";
  $str .= "#include \"socketInterface.h\"\n";
  my $port = getPort();
  my $name = getName();
  my $Uname = "\U$name\E";
  $str .= "\n";
  $str .= "#define ".$Uname."_PORT ".$port."\n";
  $str .= "\n";
  $str .= "enum ".$name."_functions_enums {\n";
#  foreach $funcName (keys(%functionFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $funcName = "";
    if (getNext(\$funcName, \%functionFields, $hashCount) == 0) {
      last;
    }
    my $Type = getFuncType($funcName);
    $str .= "  f_".$funcName." = ".$Type.",\n";
    $str .= "  f_".$funcName."_send = (f_".$funcName." << 24),\n";
    $str .= "  f_".$funcName."_reply = (f_".$funcName." << 24 | 0x00010000),\n";
    $str .= "  f_".$funcName."_reply_error = (f_".$funcName." << 24 | 0x00020000),\n";
  }
  $str .= "};\n";
  $str .= "\n";
  $str .= "enum ".$name."structures_enums {\n";
#  foreach $structureName (keys(%structureFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $structureName = "";
    if (getNext(\$structureName, \%structureFields, $hashCount) == 0) {
      last;
    }
    $str .= "  e_".$structureName." = ".getStructType($structureName).",\n";
  }
  $str .= "};\n";
  $str .= "\n";
  $str .= "\n";
#  foreach $structureName (keys(%structureFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $structureName = "";
    if (getNext(\$structureName, \%structureFields, $hashCount) == 0) {
      last;
    }
    $str .= "enum ".$structureName."_enums {\n";
    my $fieldNameArrayRef = getFields($structureName);
    my $fieldTypeArrayRef = getFieldTypes($structureName);
    my $fieldTValArrayRef = getFieldTVals($structureName);
    my $numFields = getStructNumFields($structureName);
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $T = $$fieldTValArrayRef[$j];
      $str .= "  e_".$structureName."_".$fieldName." = ".$T.",\n";
    }
    $str .= "};\n";
  }
  $str .= "\n";
  $str .= "  /* Structure Definitions */\n";
  $str .= "\n";
#  foreach $structureName (keys(%structureFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $structureName = "";
    if (getNext(\$structureName, \%structureFields, $hashCount) == 0) {
      last;
    }
    $str .= "struct ".$structureName." {\n";
    $str .= " public:\n";
    my $fieldNameArrayRef = getFields($structureName);
    my $fieldTypeArrayRef = getFieldTypes($structureName);
    my $fieldTValArrayRef = getFieldTVals($structureName);
    my $numFields = getStructNumFields($structureName);

    # Constructor
    $str .= "  ".$structureName."() {";
    my $destructorStr = "";
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $T = $$fieldTValArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type eq "int") {
	$str .= $fieldName." = 0;";
      }
      elsif ($type eq "uint") {
	$str .= $fieldName." = 0;";
      }
      elsif ($type eq "float") {
	$str .= $fieldName." = 0.0;";
      }
      elsif ($type =~ /struct/) {
	$str .= $fieldName." = NULL;";
	$destructorStr .= "delete ".$fieldName.";";
      }
    }
    $str .= "next = NULL;";
    $str .= "};\n";
    #destructor
    $str .= "  ~".$structureName."() {".$destructorStr." delete next;};\n";

    $str .= " public:\n";
    $str .= "  void print(char *pre);\n";
    $str .= "  void printBuff(char *&buff, char *pre);\n";
    $str .= "  int bufferSize();\n";
    $str .= "  void serialize(int T, char *&buffer);\n";
    $str .= "  static int deserialize(".$structureName." *&param, int T, char *buffer, int inLength);\n";
    $str .= "\n";
    $str .= " public:\n";
    #fields
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $T = $$fieldTValArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type eq "uint") {
	$str .= "    unsigned int ".$fieldName.";\n";
      }
      elsif ($type eq "string") {
	$str .= "    StringSock ".$fieldName.";\n";
      }
      elsif ($type =~ /struct/) {
	$str .= "    ".$type." *".$fieldName.";\n";
      }
      else {
	$str .= "    ".$type." ".$fieldName.";\n";
      }
    }
    $str .= "    struct ".$structureName." *next;\n";
    $str .= "};\n";
  }
  $str .= "\n";
  $str .= "class ".$className.": public Connection\n";
  $str .= "{\n";
  $str .= " public:\n";
  $str .= "  ".$className."(Connections *c, int clientOrServer, int fd, char *name = NULL) : Connection(c, clientOrServer, fd, name) {};\n";
  $str .= "  ".$className."(Connections *c, int clientOrServer, char *name = NULL) : Connection(c, clientOrServer, name) {};\n";
  $str .= "\n";
  $str .= " private:\n";
#  foreach $funcName (keys(%functionFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $funcName = "";
    if (getNext(\$funcName, \%functionFields, $hashCount) == 0) {
      last;
    }
    my $outputArrayCount = getFuncOutputArrayCount($funcName);
    my $outputArrayRef = getFuncOutputArray($funcName);

    $str .= "  int ".$funcName."ReceiveProcess(int tid, char *&inbuffer, int &len);\n";
    $str .= "  void ".$funcName."ReplyProcess(int tid);\n";
    my $callStr = "";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($i > 0) {
	$callStr .= ", ";
      }
      $callStr .= "  ".$$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_p";
    }
  }
  $str .= "  \n";
  $str .= " public:\n";
  $str .= "  virtual void processTLV();\n";
  $str .= "\n";
#  foreach $funcName (keys(%functionFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $funcName = "";
    if (getNext(\$funcName, \%functionFields, $hashCount) == 0) {
      last;
    }
    my $fieldsRef = $functionFields{$funcName};
    my $Type = getFuncType($funcName);
    my $inputArrayCount = getFuncInputArrayCount($funcName);
    my $inputArrayRef = getFuncInputArray($funcName);
    my $outputArrayCount = getFuncOutputArrayCount($funcName);
    my $outputArrayRef = getFuncOutputArray($funcName);
    my $callStr = "";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      if ($i != 0) {
	$callStr .= ", ";
      }
      $callStr .= $$inputArrayRef[$i]." *".$$inputArrayRef[$i]."_p";
    }
    if ($callStr eq "") {
      $str .= "  int  ".$funcName."TX(int tid = 0); // Default is non-blocking\n";
    }
    else {
      $str .= "  int  ".$funcName."TX(".$callStr.", int tid = 0); // Default is non-blocking\n";
    }
    my $first = 1;
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($inputArrayCount == 0 && $first) {
	$first = 0;
	$callStr .= $$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_out_p";
      }
      else {
	$callStr .= ", ".$$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_out_p";
      }
    }
    if ($callStr eq "") {
      $callStr .= "int timeout = 0";
    }
    else {
      $callStr .= ", int timeout = 0";
    }
    $str .= "  int  ".$funcName."Blocking(".$callStr."); // default = blocking\n";
    $callStr = "";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      if ($i != 0) {
	$callStr .= ", ";
      }
      $callStr .= $$inputArrayRef[$i]." *".$$inputArrayRef[$i]."_p";
    }
    $first = 1;
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($inputArrayCount == 0 && $first) {
	$first = 0;
	$callStr .= $$outputArrayRef[$i]." &out_".$$outputArrayRef[$i];
      }
      else {
	$callStr .= ", ".$$outputArrayRef[$i]." &out_".$$outputArrayRef[$i];
      }
    }
    $str .= "  virtual int  ".$funcName."RX(".$callStr.") { return 1;};\n";
    $callStr = "";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($i != 0) {
	$callStr .= ", ";
      }
      $callStr .= $$outputArrayRef[$i]." *out_".$$outputArrayRef[$i];
    }
    $str .= "  virtual void  ".$funcName."CB(".$callStr.") {};\n";
    $str .= "  virtual void  ".$funcName."CBError() {};\n";
    $str .= "\n";
  }
  $str .= "\n";
  $str .= "};\n";
  $str .= "\n";
  $str .= "#endif\n";
  $str .= "\n";
  print SESAME $str;
  print SESAME "\n";
  close(SESAME);
}


sub printConnectionBaseCppFile {
  my $fileName = getFileName();
  my $className = getFileName();
  $fileName .= ".cpp";
  print "Wrinting file $fileName\n";
  open(SESAME, ">".$fileName) or die "Could not open output file $fileName";
  my $str = "";
  $str .= "/*\n";
  $str .= "  ##################################################################################\n";
  $str .= "  #\n";
  $str .= "  #  $fileName\n";
  $str .= "  #\n";
  $str .= "  #    This is a generated file.  The functions defined in this file are primarily \n";
  $str .= "  #    used for serializing and de-serializing data structures generated from the IDL.\n";
  $str .= "  #\n";
  $str .= "  #    This distribution makes asynchronous client/server communication possible \n";
  $str .= "  #    using TLVs (Type Length Value). It achieves that letting the user define\n";
  $str .= "  #    an IDL which serves as input to a code generator.  The code generated is linked in\n";
  $str .= "  #    the client and the server program.\n";
  $str .= "  #\n";
  $str .= "  #    Copyright (C) <2005>  <Sanjay Jaiman> email: sanjayjaiman\@yahoo.com\n";
  $str .= "  #\n";
  $str .= "  #    This program is free software; you can redistribute it and/or modify\n";
  $str .= "  #    it under the terms of the GNU General Public License as published by\n";
  $str .= "  #    the Free Software Foundation; either version 2 of the License, or\n";
  $str .= "  #    (at your option) any later version.\n";
  $str .= "  #\n";
  $str .= "  #    This program is distributed in the hope that it will be useful,\n";
  $str .= "  #    but WITHOUT ANY WARRANTY; without even the implied warranty of\n";
  $str .= "  #    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n";
  $str .= "  #    GNU General Public License for more details.\n";
  $str .= "  #\n";
  $str .= "  #    You should have received a copy of the GNU General Public License\n";
  $str .= "  #    along with this program; if not, write to the Free Software\n";
  $str .= "  #    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n";
  $str .= "  #\n";
  $str .= "  ##################################################################################\n";
  $str .= "*/\n";
  $str .= "#include \"".$className.".h\"\n";
  $str .= "\n";
  my $structureNamesRef = getStructures();

#  foreach $structureName (keys(%structureFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $structureName = "";
    if (getNext(\$structureName, \%structureFields, $hashCount) == 0) {
      last;
    }
    my $numOfInts = getNumOfTypeInStruct($structureName, "int");
    my $numOfFloats = getNumOfTypeInStruct($structureName, "float");
    my $numOfStrings = getNumOfTypeInStruct($structureName, "string");
    my $numOfIpaddresses = getNumOfTypeInStruct($structureName, "ipaddress");
    my $numFields = getStructNumFields($structureName);
    my $fieldNameArrayRef = getFields($structureName);
    my $fieldTypeArrayRef = getFieldTypes($structureName);
    my $fieldTValArrayRef = getFieldTVals($structureName);
    my $numNums = $numOfInts + $numOfFloats + $numOfIpaddresses;

    $str .= "// print function of struct ".$structureName."\n";
    $str .= "void ".$structureName."::printBuff(char *&buff, char *pre)\n";
    $str .= "{\n";
    $str .= "  char tmp[80];\n";
    $str .= "  tmp[0] = \'\\0\';\n";
    $str .= "  strcpy(tmp, pre);\n";
    $str .= "  strcat(tmp, \"  \");\n";
    $str .= "  buff += sprintf (buff, \"\%s".$structureName."::\\n\", pre);\n";
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type eq "int" || $type eq "uint") {
	$str .= "  buff += sprintf (buff, \"\%s".$fieldName."=\%d;\\n\", pre, ".$fieldName.");\n";
      }
      elsif ($type eq "float") {
	$str .= "  buff += sprintf (buff, \"\%s".$fieldName."=\%f;\\n\", pre, ".$fieldName.");\n";
      }
      elsif ($type eq "ipaddress") {
	$str .= "  buff += sprintf (buff, \"\%s".$fieldName."=\%s;\\n\", pre, ".$fieldName.".charVal());\n";
      }
      elsif ($type eq "string") {
	$str .= "  buff += sprintf (buff, \"\%s".$fieldName."=\"\, pre); buff += ".$fieldName.".sprint(buff); buff += sprintf(buff, \";\\n\");\n";
      }
    }
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type =~ /struct/) {
	$str .= "  if (".$fieldName.") ".$fieldName."->printBuff(buff, tmp);\n";
      }
    }
    $str .= "  if(next) next->printBuff(buff, tmp);\n";
    $str .= "}\n";

    $str .= "\n";
    $str .= "void ".$structureName."::print(char *pre)\n";
    $str .= "{\n";
    $str .= "  char tmp[80];\n";
    $str .= "  tmp[0] = \'\\0\';\n";
    $str .= "  strcpy(tmp, pre);\n";
    $str .= "  strcat(tmp, \"  \");\n";
    $str .= "  printf (\"\%s".$structureName."::\\n\", pre);\n";
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type eq "int" || $type eq "uint") {
	$str .= "  printf (\"\%s".$fieldName."=\%d;\\n\", pre, ".$fieldName.");\n";
      }
      elsif ($type eq "float") {
	$str .= "  printf (\"\%s".$fieldName."=\%f;\\n\", pre, ".$fieldName.");\n";
      }
      elsif ($type eq "ipaddress") {
	$str .= "  printf (\"\%s".$fieldName."=\%s;\\n\", pre, ".$fieldName.".charVal());\n";
      }
      elsif ($type eq "string") {
	$str .= "  printf (\"\%s".$fieldName."=\"\, pre); ".$fieldName.".print(); printf(\";\\n\");\n";
      }
    }
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type =~ /struct/) {
	$str .= "  if (".$fieldName.") ".$fieldName."->print(tmp);\n";
      }
    }
    $str .= "  if(next) next->print(tmp);\n";
    $str .= "}\n";
    $str .= "// TLV encoding functions for struct ".$structureName."\n";
    $str .= "int ".$structureName."::bufferSize()\n";
    $str .= "{\n";
    $str .= "  // numints + (numints + thisstuct + numstrings) * 2  (2 for TL)\n";
    $str .= "  int numOfInts = ".$numNums." + (".$numNums." + 1 + ".$numOfStrings.") * 2;\n";
    $str .= "  int sum = 0;\n";
    $str .= "  sum += (numOfInts) * sizeof(int);\n";
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type =~ /struct/) {
	$str .= "  sum += ((".$fieldName.") ? ".$fieldName."->bufferSize() : 0);\n";
      }
      elsif ($type eq "string") {
	$str .= "  sum += ".$fieldName.".Length;\n";
      }
    }
    $str .= "  sum += ((next) ? next->bufferSize() : 0);\n";
    $str .= "  return sum;\n";
    $str .= "}\n";
    $str .= "void ".$structureName."::serialize(int T, char *&buffer)\n";
    $str .= "{\n";
    $str .= "  int sizeofValBuffer = bufferSize() - TLSIZE;\n";
    $str .= "  // Put in the TL\n";
    $str .= "  tlv::encodeTL(buffer, T, sizeofValBuffer);\n";
    $str .= "  // Put in the fields\n";
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type eq "int") {
	$str .= "  tlv::putIntTLV(buffer, e_".$structureName."_".$fieldName.", ".$fieldName.");\n";
      }
      elsif ($type eq "uint") {
	$str .= "  tlv::putUIntTLV(buffer, e_".$structureName."_".$fieldName.", ".$fieldName.");\n";
      }
      elsif ($type eq "float") {
	$str .= "  tlv::putFloatTLV(buffer, e_".$structureName."_".$fieldName.", ".$fieldName.");\n";
      }
      elsif ($type eq "ipaddress") {
	$str .= "  tlv::putIpaddressTLV(buffer, e_".$structureName."_".$fieldName.", ".$fieldName.");\n";
      }
      elsif ($type eq "string") {
	$str .= "  tlv::putStringTLV(buffer, e_".$structureName."_".$fieldName.", ".$fieldName.");\n";
      }
      elsif ($type =~ /struct/) {
	$str .= "  if (".$fieldName.")\n";
	$str .= "    ".$fieldName."->serialize(e_".$structureName."_".$fieldName.", buffer); // Serialize struct with type field defined in this struct\n";
      }
    }
    $str .= "  if (next) \n";
#    $str .= "    next->serialize(e_".$structureName.", buffer);\n";  -- Used to be like this
    $str .= "    next->serialize(T, buffer);\n";
    $str .= "}\n";
    $str .= "int ".$structureName."::deserialize(".$structureName." *&param, int T, char *buffer, int inLength)\n";
    $str .= "{\n";
    $str .= "  int type, length;\n";
    $str .= "  while (inLength > 0) {\n";
    $str .= "    if (!tlv::decodeTL(buffer, inLength, type, length))\n";
    $str .= "      break;\n";
    $str .= "    if (T == type) {\n";
    $str .= "      // Allocate space if needed\n";
    $str .= "      if (!param) param = new ".$structureName."();\n";
    $str .= "      // decode the rest\n";
    $str .= "      // \"buffer\" points to val buffer\n";
    for (my $j = 0; $j < $numFields; $j++) {
      my $fieldName = $$fieldNameArrayRef[$j];
      my $type = $$fieldTypeArrayRef[$j];
      if ($type eq "int") {
	$str .= "      if (!tlv::decodeINT(buffer, inLength, e_".$structureName."_".$fieldName.", param->".$fieldName."))\n";
	$str .= "         return 0;\n";
      }
      elsif ($type eq "uint") {
	$str .= "      if (!tlv::decodeUINT(buffer, inLength, e_".$structureName."_".$fieldName.", param->".$fieldName."))\n";
	$str .= "         return 0;\n";
      }
      elsif ($type eq "float") {
	$str .= "      if (!tlv::decodeFLOAT(buffer, inLength, e_".$structureName."_".$fieldName.", param->".$fieldName."))\n";
	$str .= "         return 0;\n";
      }
      elsif ($type eq "ipaddress") {
	$str .= "      if (!tlv::decodeIpaddress(buffer, inLength, e_".$structureName."_".$fieldName.", param->".$fieldName."))\n";
	$str .= "         return 0;\n";
      }
      elsif ($type eq "string") {
	$str .= "      if (!tlv::decodeSTRING(buffer, inLength, e_".$structureName."_".$fieldName.", param->".$fieldName."))\n";
	$str .= "         return 0;\n";
      }
      elsif ($type =~ /struct/) {
	$_ = $type;
	my @typeWords = split;
	my $structType = $typeWords[1];
	$str .= "      if (!".$structType."::deserialize(param->".$fieldName.", e_".$structureName."_".$fieldName.", buffer, length)) return 0;\n";
      }
    }
    $str .= "      // attempt to get the next one starting at buffer(start of this TLV) length = length of this TLV\n";
    $str .= "      if (!deserialize(param->next, T, buffer, length))\n";
    $str .= "          return 0;\n";
    $str .= "      break;\n";
    $str .= "    }\n";
    $str .= "    buffer += length;  // For Val\n";
    $str .= "    inLength -= length;\n";
    $str .= "  }\n";
    $str .= "  return 1;\n";
    $str .= "}\n";
  }
  $str .= "\n";
  $str .= "//////////////////////////////////////////////\n";
  $str .= "//      Runtime TLV dispatch functions      //\n";
  $str .= "//////////////////////////////////////////////\n";
  $str .= "\n";
  $str .= "void ".$className."::processTLV()\n";
  $str .= "{\n";
  $str .= "  int T = receiveTlv.getType();\n";
  $str .= "  int tid = T & 0x0000ffff;\n";
  $str .= "  char *sendBuffer = 0;\n";
  $str .= "  int  sendLen;\n";
  $str .= "  int  dataToSend = 0;   //\n";
  $str .= "\n";
  $str .= "  T ^= tid;  // Mask off the tansaction id to get the type\n";
  $str .= "  // cout << \"Data received on socket\" << getFd() << endl;\n";
  $str .= "  // Print the received TLV\n";
  $str .= "  // receiveTlv.print();\n";
  $str .= "\n";
  $str .= "  // Based on the type field of receiveTlv call one of Receive/reply/error functions\n";
  $str .= "  transaction *p = transaction::findtid(tid);\n";
  $str .= "  if (p) {\n";
  $str .= "    p->wakeup();\n";
  $str .= "    p->lastTypeReceived = T;  // To check for error\n";
  $str .= "    p->postWait(); // Wait for other thread to finish processing (blocking)\n";
  $str .= "    //\n";
  $str .= "    // It is important to delete the transaction control block here as opposed to in\n";
  $str .= "    // \"<function>Blocking\" as by the time this thread wakes up, \"trans\" may have already\n";
  $str .= "    // been deleted by the \"<function>Blocking\" and hence causing a SEGV in \"postWakeup\".\n";
  $str .= "    //\n";
  $str .= "    delete p;\n";
  $str .= "    return;\n";
  $str .= "  }\n";
  $str .= "\n";
  $str .= "  switch (T) {\n";
#  foreach $funcName (keys(%functionFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $funcName = "";
    if (getNext(\$funcName, \%functionFields, $hashCount) == 0) {
      last;
    }
    $str .= "  case f_".$funcName."_send:\n";
    $str .= "    dataToSend = ".$funcName."ReceiveProcess(tid, sendBuffer, sendLen);\n";
    $str .= "    break;\n";
    $str .= "  case f_".$funcName."_reply:\n";
    $str .= "    ".$funcName."ReplyProcess(tid);\n";
    $str .= "    break;\n";
    $str .= "  case f_".$funcName."_reply_error:\n";
    $str .= "    break;\n";
  }
  $str .= "  default:\n";
  $str .= "    break;\n";
  $str .= "  }\n";
  $str .= "  if (dataToSend) {\n";
  $str .= "    Send(sendBuffer, sendLen);\n";
  $str .= "  }\n";
  $str .= "  delete(sendBuffer);\n";
  $str .= "}\n";
  $str .= "\n";
  $str .= "// Receive functions deserialize the input TLV / call the RX callback fn / build the reply TLV\n";
  $str .= "// These functions allocate spcae in the pointer ref. provided - Caller has to free the buffer.\n";
  $str .= "\n";
#  foreach $funcName (keys(%functionFields)) {
  for (my $hashCount = 0; ; $hashCount++) {
    my $funcName = "";
    if (getNext(\$funcName, \%functionFields, $hashCount) == 0) {
      last;
    }
    my $fieldsRef = $functionFields{$funcName};
    my $Type = getFuncType($funcName);
    my $inputArrayCount = getFuncInputArrayCount($funcName);
    my $inputArrayRef = getFuncInputArray($funcName);
    my $outputArrayCount = getFuncOutputArrayCount($funcName);
    my $outputArrayRef = getFuncOutputArray($funcName);
    $str .= "int ".$className."::".$funcName."ReceiveProcess(int tid, char *&inbuffer, int &len)\n";
    $str .= "{\n";
    $str .= "  char *V = receiveTlv.getVal();\n";
    $str .= "  (void)V;\n";
    $str .= "  int L = receiveTlv.getLength();\n";
    $str .= "  char *p = NULL;\n";
    $str .= "  int T = 0;  // For sending back\n";
    $str .= "\n";
    $str .= "  if (L < 0) return 0; // 0 is OK in case of no parms\n";
    $str .= "\n";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      $str .= "  ".$$inputArrayRef[$i]." *".$$inputArrayRef[$i]."_p = NULL;\n";
    }
    $str .= "\n";
    $str .= "  // Temp space for RX function\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  ".$$outputArrayRef[$i]." out_".$$outputArrayRef[$i].";\n";
    }
    $str .= "  int buffSize = 0;\n";
    $str .= "\n";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      $str .= "  if (!".$$inputArrayRef[$i]."::deserialize(".$$inputArrayRef[$i]."_p, e_".$$inputArrayRef[$i].", V, L) || ".$$inputArrayRef[$i]."_p == NULL) goto error;\n";
    }
    $str .= "\n";
    my $RXcallStr = "";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      if ($i != 0) {
	$RXcallStr .= ", ";
      }
      $RXcallStr .= $$inputArrayRef[$i]."_p";
    }
    my $first = 1;
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($inputArrayCount == 0 && $first) {
	$first = 0;
	$RXcallStr .= "out_".$$outputArrayRef[$i];
      }
      else {
	$RXcallStr .= ",  out_".$$outputArrayRef[$i];
      }
    }
    $str .= "  if (!".$funcName."RX(".$RXcallStr.")) goto error;\n";
    $str .= "\n";
    $str .= "  // Allocate buffer for sending\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  buffSize += out_".$$outputArrayRef[$i].".bufferSize();\n";
    }
    $str .= "  p = inbuffer = new char[buffSize + TLSIZE]; // TL + buffer\n";
    $str .= "  // Put in the TL\n";
    $str .= "  T = f_".$funcName."_reply | tid;\n";
    $str .= "  tlv::encodeTL(p, T, buffSize);\n";
    $str .= "\n";
    $str .= "  // Put in tke V\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  out_".$$outputArrayRef[$i].".serialize(e_".$$outputArrayRef[$i].", p); // struct\n";
    }
    $str .= "  len = buffSize + TLSIZE;\n";
    $str .= "  goto exit;\n";
    $str .= "\n";
    $str .= " error:\n";
    $str .= "  p = inbuffer = new char[TLSIZE];  // TL\n";
    $str .= "  // Put in the TL - No V\n";
    $str .= "  T = f_".$funcName."_reply_error | tid;\n";
    $str .= "  len = tlv::encodeTL(p, T, 0);\n";
    $str .= "\n";
    $str .= " exit:\n";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      $str .= "  delete ".$$inputArrayRef[$i]."_p;\n";
    }
    $str .= "  return len;\n";
    $str .= "}\n";
    $str .= "\n";
    $str .= "// Reply function :: deserialize the output TLV / call the action CB callback fn\n";
    $str .= "void ".$className."::".$funcName."ReplyProcess(int tid)\n";
    $str .= "{\n";
    $str .= "  char *V = receiveTlv.getVal();\n";
    $str .= "  int L = receiveTlv.getLength();\n";
    $str .= "\n";
    $str .= "  if ((L < 0) || (!V && L > 0)) {\n";
    $str .= "    ".$funcName."CBError();\n";
    $str .= "    return;\n";
    $str .= "  }\n";
    $str .= "  // For CB function\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  ".$$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_p = NULL;\n";
    }
    $str .= "\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  ".$$outputArrayRef[$i]."::deserialize(".$$outputArrayRef[$i]."_p, e_".$$outputArrayRef[$i].", V, L);\n";
    }
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  if (!".$$outputArrayRef[$i]."_p) {\n";
      $str .= "    ".$funcName."CBError();\n";
      $str .= "    return;\n";
      $str .= "  }\n";
    }
    my $callStr = "";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($i != 0) {
	$callStr .= ", ";
      }
      $callStr .= $$outputArrayRef[$i]."_p";
    }
    $str .= "  ".$funcName."CB(". $callStr.");\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  delete ".$$outputArrayRef[$i]."_p;\n";
    }
    $str .= "}\n";
    $str .= "\n";
    $callStr = "";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($i > 0) {
	$callStr .= ", ";
      }
      $callStr .= "  ".$$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_p";
    }
    $str .= "\n";
    $str .= "// TX function :: Used as input point for initiating RPC\n";
    $callStr = "";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      if ($i != 0) {
	$callStr .= ", ";
      }
      $callStr .= $$inputArrayRef[$i]." *".$$inputArrayRef[$i]."_p";
    }
    if ($callStr ne "") {
      $str .= "int  ".$className."::".$funcName."TX(".$callStr.", int tid)\n";
    }
    else {
      $str .= "int  ".$className."::".$funcName."TX(int tid)\n";
    }
    $str .= "{\n";
    $str .= "  if (!isConnected()) return 0;\n";
    $str .= "  int buffSize = 0;\n";
    $str .= "  int T = f_".$funcName."_send; // Type val to be sent\n";
    $str .= "\n";
    $str .= "  if (tid) { // if tid == non-zero => blocking call\n";
    $str .= "    T |= tid;  // Put the trans id in the Type field\n";
    $str .= "  }\n";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      $str .= "  if (!".$$inputArrayRef[$i]."_p) return 0;\n";
    }
    $str .= "\n";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      $str .= "  buffSize += ".$$inputArrayRef[$i]."_p->bufferSize();\n";
    }
    $str .= "  char data[buffSize];\n";
    $str .= "  char *p = data;\n";
    $str .= "  (void)p;\n";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      $str .= "  ".$$inputArrayRef[$i]."_p->serialize(e_".$$inputArrayRef[$i].", p);\n";
    }
    $str .= "  tlv TLV(T, buffSize, data);\n";
    $str .= "  if (!sendTlv(TLV)) {\n";
    $str .= "    return 0;\n";
    $str .= "  }\n";
    $str .= "  return 1;  // success\n";
    $str .= "}\n";
    my $first = 1;
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($inputArrayCount == 0 && $first) {
	$first = 0;
	$callStr .= $$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_out_p";
      }
      else {
	$callStr .= ", ".$$outputArrayRef[$i]." *".$$outputArrayRef[$i]."_out_p";
      }
    }
    if ($callStr eq "") {
      $callStr .= "int timeout = 0";
    }
    else {
      $callStr .= ", int timeout = 0";
    }
    $str .= "int  ".$className."::".$funcName."Blocking(".$callStr.")\n";
    $str .= "{\n";
    $callStr = "";
    for (my $i = 0; $i < $inputArrayCount; $i++) {
      if ($i > 0) {
	$callStr .= ", ";
      }
      $callStr .= $$inputArrayRef[$i]."_p";
    }
    my $callStr1 = "";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      if ($i > 0) {
	$callStr1 .= ", ";
      }
      $callStr1 .= $$outputArrayRef[$i]."_out_p";
    }
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "  if (!".$$outputArrayRef[$i]."_out_p) return 0; // check if output is valid\n";
    }
    $str .= "  transaction *trans = transaction::getTransaction(timeout); //allocates tranaction Control Block\n";
    $str .= "  if (!trans) return 0;\n";
    $str .= "  int tid = trans->getTid();\n";
    $str .= "  if (!tid) {\n";
    $str .= "    delete trans;\n";
    $str .= "    return 0;\n";
    $str .= "  };\n";
    $str .= "  int returnCode = 1;\n";
    if ($callStr eq "") {
      $str .= "  if (".$funcName."TX(tid)) {\n";
    }
    else {
      $str .= "  if (".$funcName."TX(".$callStr.", tid)) {\n";
    }

    $str .= "    if (!trans->timedWait()) {\n";
    $str .= "      delete trans;  // trans deleted here as calling process doesn't delete this block\n";
    $str .= "      returnCode = 0;\n";
    $str .= "    }\n";
    $str .= "    else {\n";
    $str .= "      //We got here because we got we were woken up by \"processTLV\" called \"wakeup()\" on the tarns\n";
    $str .= "      //Therefore after this function returns, processTLV will delete the trans control block\n";
    $str .= "      char *V = receiveTlv.getVal();\n";
    $str .= "      int L = receiveTlv.getLength();\n";
    $str .= "      int T = trans->lastTypeReceived;\n";
    $str .= "  \n";
    $str .= "      if ((L < 0) || (!V && L > 0)) {\n";
    $str .= "        returnCode = 0;\n";
    $str .= "      }\n";
    $str .= "      else if( T == f_".$funcName."_reply_error) {\n";
    $str .= "        returnCode = 0;\n";
    $str .= "      }\n";
    $str .= "      else {\n";
    for (my $i = 0; $i < $outputArrayCount; $i++) {
      $str .= "        ".$$outputArrayRef[$i]."::deserialize(".$$outputArrayRef[$i]."_out_p, e_".$$outputArrayRef[$i].", V, L);\n";
    }
    $str .= "      }\n";
    $str .= "      trans->postWakeup();  // Wake up calling thread\n";
    $str .= "      //\n";
    $str .= "      // It is important NOT to delete the transaction control block here ProcessTLV\n";
    $str .= "      // is waiting on the semaphore in this structure. We will let ProcessTLV delete\n";
    $str .= "      // the transaction control block AFTER it wakes up to avoid a SEGV\n";
    $str .= "      //\n";
    $str .= "    }\n";
    $str .= "  }\n";
    $str .= "  return returnCode;\n";
    $str .= "}\n";
    $str .= "\n";
  }
  print SESAME $str;
  print SESAME "\n";
  close(SESAME);
}

#### Main Program ####

my $numArgs = @ARGV;
$inFilename = ($numArgs) ? $ARGV[0] : "test_input";

print "Input File - $inFilename\n";

open(INPUT, $inFilename) or die "Could not open input file $inFilename";

parseFile(INPUT, SESAME);
my $name = "root";

#printVars($name, \%globalVariableTree, $numGlobalVariables);

print "closing file $inFilename\n";
close(INPUT);

#doUserIo();
buildTLVStructures();
#printStructures();
#printFunctions();
printConnectionBaseCppFile();
printConnectionBaseHFile();

